;; Generated package description from highlight-escape-sequences.el  -*- no-byte-compile: t -*-
(define-package "highlight-escape-sequences" "0.4.0.20201214.173014" "Highlight escape sequences" 'nil :commit "fae976568c04b6fe8a9f2d854c8fe23b357a6878" :authors '(("Dmitry Gutov" . "dgutov@yandex.ru") ("Pavel Matcula" . "dev.plvlml@gmail.com")) :maintainer '(("Dmitry Gutov" . "dgutov@yandex.ru") ("Pavel Matcula" . "dev.plvlml@gmail.com")) :keywords '("convenience") :url "https://github.com/dgutov/highlight-escape-sequences")
